.hl {
  background-color: var(--hl-strong);
}
span.hl {
	background-color: var(--hl-strong);
	border-width: 0;
	border-radius: 2px;
	border-style: solid;
}
div.contnr.hl,div.lbl.hl {
  background-color: var(--hl-strong);
}
div.contnr.hl {
  border-color: var(--hl-rim) ! important;
	border-width: 4px ! important;
}

span.hlbx {
	border-color: var(--hl-rim);
	border-width: 4px ! important;
	border-style: solid;
	border-radius: 6px;
  padding: 4px;
  margin: 4px;
}
.ehl {
  background-color: var(--ehl-strong);
}

:root {
	--hl-strong:        hsla( 60, 100%,  70%, 0.9  );
	--hl-rim:           hsla( 55,  80%,  50%, 1.0  );
	--ehl-strong:       hsla(240, 100%,  70%, 0.9  );
}
