@font-face {
  font-family: "Gentium Plus";
  src: local('Gentium Plus'), local('GentiumPlus'),
    url('/browser/static/fonts/GentiumPlus-R.woff') format('woff'),
    url('https://github.com/annotation/text-fabric/blob/master/tf/browser/static/fonts/GentiumPlus-R.woff?raw=true') format('woff');
}

@font-face {
  font-family: "Ezra SIL";
  src: local('Ezra SIL'), local('EzraSIL'),
    url('/browser/static/fonts/SILEOT.woff') format('woff'),
    url('https://github.com/annotation/text-fabric/blob/master/tf/browser/static/fonts/SILEOT.woff?raw=true') format('woff');
}

@font-face {
  font-family: "SBL Hebrew";
  src: local('SBL Hebrew'), local('SBLHebrew'),
    url('/browser/static/fonts/SBL_Hbrw.woff') format('woff'),
    url('https://github.com/annotation/text-fabric/blob/master/tf/browser/static/fonts/SBL_Hbrw.woff?raw=true') format('woff');
}

@font-face {
  font-family: "Estrangelo Edessa";
  src: local('Estrangelo Edessa'), local('EstrangeloEdessa');
    url('/browser/static/fonts/SyrCOMEdessa.woff') format('woff'),
    url('https://github.com/annotation/text-fabric/blob/master/tf/browser/static/fonts/SyrCOMEdessa.woff?raw=true') format('woff');
}

@font-face {
  font-family: AmiriQuran;
  font-style: normal;
  font-weight: 400;
  src: local('Amiri Quran'), local('AmiriQuran'),
    url('/browser/static/fonts/AmiriQuran.woff') format('woff'),
    url('https://github.com/annotation/text-fabric/blob/master/tf/browser/static/fonts/AmiriQuran.woff?raw=true') format('woff');
}

@font-face {
  font-family: AmiriQuranColored;
  font-style: normal;
  font-weight: 400;
  src: local('Amiri Quran Colored'), local('AmiriQuranColored'),
    url('/browser/static/fonts/AmiriQuranColored.woff') format('woff'),
    url('https://github.com/annotation/text-fabric/blob/master/tf/browser/static/fonts/AmiriQuranColored.woff?raw=true') format('woff');
}

@font-face {
  font-family: "Santakku";
  src: local('Santakku'),
    url('/browser/static/fonts/Santakku.woff') format('woff'),
    url('https://github.com/annotation/text-fabric/blob/master/tf/browser/static/fonts/Santakku.woff?raw=true') format('woff');
}

@font-face {
  font-family: "SantakkuM";
  src: local('SantakkuM'),
    url('/browser/static/fonts/SantakkuM.woff') format('woff'),
    url('https://github.com/annotation/text-fabric/blob/master/tf/browser/static/fonts/SantakkuM.woff?raw=true') format('woff');
}

