/* bypassing some classical notebook settings */
div#notebook {
  line-height: unset;
}
/* neutral text */
.txtn,.txtn a:visited,.txtn a:link {
    font-family: sans-serif;
    font-size: medium;
    direction: ltr;
    unicode-bidi: embed;
    text-decoration: none;
    color: var(--text-color);
}
/* transcription text */
.txtt,.txtt a:visited,.txtt a:link {
    font-family: monospace;
    font-size: medium;
    direction: ltr;
    unicode-bidi: embed;
    text-decoration: none;
    color: var(--text-color);
}
/* source text */
.txto,.txto a:visited,.txto a:link {
    font-family: serif;
    font-size: medium;
    direction: ltr;
    unicode-bidi: embed;
    text-decoration: none;
    color: var(--text-color);
}
/* phonetic text */
.txtp,.txtp a:visited,.txtp a:link {
    font-family: Gentium, sans-serif;
    font-size: medium;
    direction: ltr;
    unicode-bidi: embed;
    text-decoration: none;
    color: var(--text-color);
}
/* original script text */
.txtu,.txtu a:visited,.txtu a:link {
    font-family: Gentium, sans-serif;
    font-size: medium;
    text-decoration: none;
    color: var(--text-color);
}
/* hebrew */
.txtu.hbo,.lex.hbo {
    font-family: "Ezra SIL", "SBL Hebrew", sans-serif;
    font-size: large;
    direction: rtl ! important;
    unicode-bidi: embed;
}
/* syriac */
.txtu.syc,.lex.syc {
    font-family: "Estrangelo Edessa", sans-serif;
    font-size: medium;
    direction: rtl ! important;
    unicode-bidi: embed;
}
/* neo aramaic */
.txtu.cld,.lex.cld {
    font-family: "CharisSIL-R", sans-serif;
    font-size: medium;
    direction: ltr ! important;
    unicode-bidi: embed;
}
/* standard arabic */
.txtu.ara,.lex.ara {
    font-family: "AmiriQuran", sans-serif;
    font-size: large;
    direction: rtl ! important;
    unicode-bidi: embed;
}
/* cuneiform */
.txtu.akk,.lex.akk {
    font-family: Santakku, sans-serif;
    font-size: large;
    direction: ltr ! important;
    unicode-bidi: embed;
}
.txtu.uga,.lex.uga {
    font-family: sans-serif;
    /*font-family: Noto Sans Ugaritic, sans-serif;*/
    font-size: large;
    direction: ltr ! important;
    unicode-bidi: embed;
}
/* greek */
.txtu.grc,.lex.grc a:link {
    font-family: Gentium, sans-serif;
    font-size: medium;
    direction: ltr ! important;
    unicode-bidi: embed;
}
a:hover {
    text-decoration: underline | important;
    color: #0000ff | important;
}
.ltr {
    direction: ltr ! important;
    unicode-bidi: embed;
}
.rtl {
    direction: rtl ! important;
    unicode-bidi: embed;
}
.ubd {
    unicode-bidi: embed;
}
.col {
   display: inline-block;
}
.features {
    font-family: monospace;
    font-size: medium;
    font-weight: bold;
    color: var(--features);
    display: flex;
    flex-flow: column nowrap;
    justify-content: flex-start;
    align-items: flex-start;
    align-content: flex-start;
    padding: 2px;
    margin: 2px;
    direction: ltr;
    unicode-bidi: embed;
    border: var(--meta-width) solid var(--meta-color);
    border-radius: var(--meta-width);
}
.features div,.features span {
    padding: 0;
    margin: -2px 0;
}
.features .f {
    font-family: sans-serif;
    font-size: small;
    font-weight: normal;
    color: #5555bb;
}
.features .xft {
  color: #000000;
  background-color: #eeeeee;
  font-size: medium;
  margin: 2px 0px;
}
.features .xft .f {
  color: #000000;
  background-color: #eeeeee;
  font-size: small;
  font-weight: normal;
}
.tfsechead {
    font-family: sans-serif;
    font-size: small;
    font-weight: bold;
    color: var(--tfsechead);
    unicode-bidi: embed;
    text-align: start;
}
.structure {
    font-family: sans-serif;
    font-size: small;
    font-weight: bold;
    color: var(--structure);
    unicode-bidi: embed;
    text-align: start;
}
.comments {
    display: flex;
    justify-content: flex-start;
    align-items: flex-start;
    align-content: flex-start;
    flex-flow: column nowrap;
}
.nd, a:link.nd {
    font-family: sans-serif;
    font-size: small;
    color: var(--node);
    vertical-align: super;
    direction: ltr ! important;
    unicode-bidi: embed;
}
.nde, a:link.nde {
    font-family: sans-serif;
    font-size: small;
    color: var(--node);
    direction: ltr ! important;
    unicode-bidi: embed;
}
.etf {
    font-size: normal;
    border-radius: 0.2rem;
    border: 1pt solid white;
    padding: 0 0.2rem ! important;
    margin: 0 0.2rem ! important;
}
.etfx {
    font-size: x-large;
}
.lex {
  color: var(--lex-color);;
}
#colormapplus, #colormapmin, .ecolormapmin {
  font-weight: bold;
  border-radius: 0.1rem;
  background-color: #eeeeff;
  padding: 0 1rem;
  margin: 0 1rem;
}
.clr {
  font-style: italic;
  font-size: small;
}
.clmap,.eclmap {
  padding: 0;
}
.children,.children.ltr {
    display: flex;
    border: 0;
    background-color: #ffffff;
    justify-content: flex-start;
    align-items: flex-start;
    align-content: flex-start;
}
.children.stretch {
    align-items: stretch;
}
.children.hor {
    flex-flow: row nowrap;
}
.children.hor.wrap {
    flex-flow: row wrap;
}
.children.ver {
    flex-flow: column nowrap;
}
.children.ver.wrap {
    flex-flow: column wrap;
}
.contnr {
    width: fit-content;
    display: flex;
    justify-content: flex-start;
    align-items: flex-start;
    align-content: flex-start;
    flex-flow: column nowrap;
    background: #ffffff none repeat scroll 0 0;
    padding:  10px 2px 2px 2px;
    margin: 16px 2px 2px 2px;
    border-style: solid;
    font-size: small;
}
.contnr.trm {
    background-attachment: local;
}
.contnr.cnul {
    padding:  0;
    margin: 0;
    border-style: solid;
    font-size: xx-small;
}
.contnr.cnul,.lbl.cnul {
    border-color: var(--border-color-nul);
    border-width: var(--border-width-nul);
    border-radius: var(--border-width-nul);
}
.contnr.c0,.lbl.c0 {
    border-color: var(--border-color0);
    border-width: var(--border-width0);
    border-radius: var(--border-width0);
}
.contnr.c1,.lbl.c1 {
    border-color: var(--border-color1);
    border-width: var(--border-width1);
    border-radius: var(--border-width1);
}
.contnr.c2,.lbl.c2 {
    border-color: var(--border-color2);
    border-width: var(--border-width2);
    border-radius: var(--border-width2);
}
.contnr.c3,.lbl.c3 {
    border-color: var(--border-color3);
    border-width: var(--border-width3);
    border-radius: var(--border-width3);
}
.contnr.c4,.lbl.c4 {
    border-color: var(--border-color4);
    border-width: var(--border-width4);
    border-radius: var(--border-width4);
}
span.plain {
    /*display: inline-block;*/
    display: inline-flex;
    flex-flow: row wrap;
    white-space: pre-wrap;
}
span.break {
  flex-basis: 100%;
  height: 0;
}
.plain {
    background-color: #ffffff;
}
.plain.l,.contnr.l,.contnr.l>.lbl {
    border-left-style: dotted
}
.plain.r,.contnr.r,.contnr.r>.lbl {
    border-right-style: dotted
}
.plain.lno,.contnr.lno,.contnr.lno>.lbl {
    border-left-style: none
}
.plain.rno,.contnr.rno,.contnr.rno>.lbl {
    border-right-style: none
}
.plain.l {
    padding-left: 4px;
    margin-left: 2px;
    border-width: var(--border-width-plain);
}
.plain.r {
    padding-right: 4px;
    margin-right: 2px;
    border-width: var(--border-width-plain);
}
.lbl {
    font-family: monospace;
    margin-top: -24px;
    margin-left: 20px;
    background: #ffffff none repeat scroll 0 0;
    padding: 0 6px;
    border-style: solid;
    display: block;
    color: var(--label)
}
.lbl.trm {
    background-attachment: local;
    margin-top: 2px;
    margin-left: 2px;
    padding: 2px 2px;
    border-style: none;
}
.lbl.cnul {
    font-size: xx-small;
}
.lbl.c0 {
    font-size: small;
}
.lbl.c1 {
    font-size: small;
}
.lbl.c2 {
    font-size: medium;
}
.lbl.c3 {
    font-size: medium;
}
.lbl.c4 {
    font-size: large;
}
.occs, a:link.occs {
    font-size: small;
}

/* PROVENANCE */

div.prov {
	margin: 40px;
	padding: 20px;
	border: 2px solid var(--fog-rim);
}
div.pline {
	display: flex;
	flex-flow: row nowrap;
	justify-content: stretch;
	align-items: baseline;
}
div.p2line {
	margin-left: 2em;
	display: flex;
	flex-flow: row nowrap;
	justify-content: stretch;
	align-items: baseline;
}
div.psline {
	display: flex;
	flex-flow: row nowrap;
	justify-content: stretch;
	align-items: baseline;
	background-color: var(--gold-mist-back);
}
div.pname {
	flex: 0 0 5rem;
	font-weight: bold;
}
div.pval {
    flex: 1 1 auto;
}

/* KEYBOARD */
.ccoff {
  background-color: inherit;
}
.ccon {
  background-color: yellow ! important;
}
.ccon,.ccoff {
  padding: 0.2rem;
  margin: 0.2rem;
  border: 0.1rem solid var(--letter-box-border);
  border-radius: 0.1rem;
}
.ccline {
  font-size: xx-large ! important;
  font-weight: bold;
  line-height: 2em ! important;
}
/* TF header */

summary {
  /* needed to override the normalize.less
   * in the classical Jupyter Notebook
   */
  display: list-item ! important;
}

.fcorpus {
  display: flex;
  flex-flow: column nowrap;
  justify-content: flex-start;
  align-items: flex-start;
  align-content: flex-start;
  overflow: auto;
}
.frow {
  display: flex;
  flex-flow: row nowrap;
  justify-content: flex-start;
  align-items: flex-start;
  align-content: flex-start;
}
.fmeta {
  display: flex;
  flex-flow: column nowrap;
  justify-content: flex-start;
  align-items: flex-start;
  align-content: flex-start;
}
.fmetarow {
  display: flex;
  flex-flow: row nowrap;
  justify-content: flex-start;
  align-items: flex-start;
  align-content: flex-start;
}
.fmetakey {
  min-width: 8em;
  font-family: monospace;
}
.fnamecat {
  min-width: 8em;
}
.fnamecat.edge {
  font-weight: bold;
  font-style: italic;
}
.fmono {
    font-family: monospace;
}

:root {
	--node:               hsla(120, 100%,  20%, 1.0  );
	--label:              hsla(  0, 100%,  20%, 1.0  );
	--tfsechead:          hsla(  0, 100%,  25%, 1.0  );
	--structure:          hsla(120, 100%,  25%, 1.0  );
	--features:           hsla(  0,   0%,  30%, 1.0  );
  --text-color:         hsla( 60,  80%,  10%, 1.0  );
  --lex-color:          hsla(220,  90%,  60%, 1.0  );
  --meta-color:         hsla(  0,   0%,  90%, 0.7  );
  --meta-width:         3px;
  --border-color-nul:   hsla(  0,   0%,  90%, 0.5  );
  --border-color0:      hsla(  0,   0%,  90%, 0.9  );
  --border-color1:      hsla(  0,   0%,  80%, 0.9  );
  --border-color2:      hsla(  0,   0%,  70%, 0.9  );
  --border-color3:      hsla(  0,   0%,  80%, 0.8  );
  --border-color4:      hsla(  0,   0%,  60%, 0.9  );
	--letter-box-border:  hsla(  0,   0%,  80%, 0.5  );
  --border-width-nul:   2px;
  --border-width0:      2px;
  --border-width1:      3px;
  --border-width2:      4px;
  --border-width3:      6px;
  --border-width4:      5px;
  --border-width-plain: 2px;
}
