"""
# Various forms of data interchange

TF can interchange data with MQL, see `tf.convert.mql`.

It has also support for converting streams of data to TF format,
see `tf.convert.walker`.

And there is support for round trips of TF data to annotation tools and back,
see `tf.convert.recorder`.

You can *explode* plain TF files in big, non-optimised data files,
see `tf.convert.tf.explode`.
"""
