<!DOCTYPE html>
<html> 
	<head>
		<meta charset="utf-8"/>
		<meta name="viewport" content="width=device-width, initial-scale=1.0"/>
    <title>TF-{{toolkey}} {{appname}}</title>
		<meta name="application-name" content="TF Named Entities"/>
		<link rel="shortcut icon" href="/browser/static/favicon.ico">
		<link rel="stylesheet" href="/browser/static/colors.css"/>
		<link rel="stylesheet" href="/browser/static/display.css"/>
		<link rel="stylesheet" href="/browser/static/highlight.css"/>
    <link rel="stylesheet" href="/browser/static/fonts.css"/>
    <link rel="stylesheet" href="/{{toolkey}}/static/base.css"/>
    <link rel="stylesheet" href="/{{toolkey}}/static/index.css"/>
    {{css|safe}}
		<script defer src="/browser/static/jquery.js"></script>
    <script defer src="/{{toolkey}}/static/tool.js"></script>
	</head>
	<body>
		<form id="go" method="post" enctype="multipart/form-data">
      <input type="hidden" id="toolkey" value="{{toolkey}}">
      <input type="hidden" id="featurelist" value="{{featurelist}}">
      <input type="hidden" id="submitter" name="submitter" value="{{submitter}}">
      <input type="hidden" id="resetf" name="resetform" value="{{resetform}}">
      <input type="hidden" id="appname" name="appname" value="{{appname}}"/>
      <input type="hidden" id="slottype" name="slottype" value="{{slottype}}"/>
      <input type="hidden" id="buckettype" name="buckettype" value="{{buckettype}}"/>
      <div class="page">
        <div id="col1">
          <p><a href="/">back to TF browser</a></p>
          <h1>{{appname}}</h1>
          <h2>Named Entity Editing</h2>
          {{messages|safe}}
          {{tasks|safe}}
          {{caption|safe}}
          <div id="logs">
            {{logs|safe}}
          </div>
        </div>
        <div id="col2">
          <div id="entitiesq">
            <h1>Entities</h1>
            {{entityoverview|safe}}
            {{entityheaders|safe}}
            {{subtlebuttons|safe}}
            <p><input type="text" name="efind" id="efind" value="{{efind}}">
                <span class="stat" id="nentityentries"></span>
                <button type="button" id="entityclear" class="icon">❌</button>
                <input type="hidden"
                       id="activeentity"
                       name="activeentity"
                       value="{{activeentityrep}}"
                >
                <input type="hidden"
                       id="activetrigger"
                       name="activetrigger"
                       value="{{activetriggerrep}}"
                >
                {{activevalrep|safe}}
            </p>
          </div>
          <hr width="100%">
          <div id="entities">
            {{entitytable|safe}}
          </div>
        </div>
        <div id="col3">
          <div id="bucketsq">
            {{entityinit|safe}}
            <div id="appearancewidget">
              <b>Appearance</b>
              {{formattingbuttons|safe}}
            </div>
            {{filterwidget|safe}}
            {{entitytext|safe}}
            {{selectentities|safe}}
            {{modifyreport|safe}}
            {{modifyentity|safe}}
          </div>
          <hr width="100%">
          {{buckets|safe}}
        </div>
      </div>
    </form>
	</body>
</html>
